/**
 * Copyright 2010 CSIRO
 * www.csiro.au
 * 
 * Contact: Carsten.Friedrich@csiro.au
 */
package csiro.opennebula;

import java.io.BufferedInputStream;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;

import org.opennebula.client.Client;
import org.opennebula.client.OneResponse;

import csiro.opennebula.jaxb.Cluster;
import csiro.opennebula.jaxb.ClusterPool;
import csiro.opennebula.jaxb.Host;
import csiro.opennebula.jaxb.HostPool;
import csiro.opennebula.jaxb.ImageDetails;
import csiro.opennebula.jaxb.ImagePool;
import csiro.opennebula.jaxb.User;
import csiro.opennebula.jaxb.UserPool;
import csiro.opennebula.jaxb.VirtualMachine;
import csiro.opennebula.jaxb.VirtualMachinePool;
import csiro.opennebula.jaxb.VirtualNetworkDetails;
import csiro.opennebula.jaxb.VirtualNetworkPool;

/**
 * @author fri096
 * 
 */
public class OneJaxbClient extends Client {

	public final static int ALL_USERS = -2;
	public final static int CURRENT_USER = -1;

	public enum VmAction {
		SHUTDOWN("shutdown"), HOLD("hold"), RELEASE("release"), STOP("stop"), CANCEL(
				"cancel"), SUSPEND("suspend"), RESUME("resume"), RESTART(
				"restart"), FINALIZE("finalize");

		private String code;

		private VmAction(String code) {
			this.code = code;
		}

		public String getCode() {
			return code;
		}
	}

	public static final String[] HOST_STATES = { "INIT", "MONITORING",
			"MONITORED", "ERROR", "DISABLED" };

	public static final String[] VM_STATES = { "INIT", "PENDING", "HOLD",
			"ACTIVE", "STOPPED", "SUSPENDED", "DONE", "FAILED" };

	public static final String[] SHORT_VM_STATES = { "init", "pend", "hold",
			"actv", "stop", "susp", "done", "fail" };

	public static final String[] LCM_STATE = { "LCM_INIT", "PROLOG", "BOOT",
			"RUNNING", "MIGRATE", "SAVE_STOP", "SAVE_SUSPEND", "SAVE_MIGRATE",
			"PROLOG_MIGRATE", "PROLOG_RESUME", "EPILOG_STOP", "EPILOG",
			"SHUTDOWN", "CANCEL", "FAILURE", "DELETE", "UNKNOWN" };

	public static final String[] SHORT_LCM_STATES = { null, "prol", "boot",
			"runn", "migr", "save", "save", "save", "migr", "prol", "epil",
			"epil", "shut", "shut", "fail", "dele", "unkn" };

	public static final String[] IMAGE_STATE = { "INIT", "READY", "USED",
			"DISABLED" };
	public static final String[] IMAGE_TYPE = { "OS", "CDROM", "DATABLOCK" };

	public static Map<String, String> imageNameToPathMap = new HashMap<String, String>();

	public Collection<ImageDetails> getImages() throws IOException {
		List<ImageDetails> res = imagePool(ALL_USERS).getImages();
		return res;
	}

	public OneJaxbClient(String username, String secret, String endpoint)
			throws Exception {
		super(username + ":" + secret, endpoint);
	}

	protected Object jaxbCall(String cmd,
			Class<? extends Object> topLevelReturn, Object... args)
			throws IOException {
		OneResponse res = super.call(cmd, args);
		if (res.isError())
			throw new IOException(res.getErrorMessage());

		try {
			JAXBContext jc = JAXBContext.newInstance(topLevelReturn);
			Unmarshaller u = jc.createUnmarshaller();
			BufferedInputStream is = new BufferedInputStream(
					new ByteArrayInputStream(res.getMessage().getBytes("UTF-8")));
			return u.unmarshal(is);
		} catch (JAXBException e) {
			throw new IOException(e.getMessage(), e);
		}
	}

	protected OneResponse jaxbCall(String cmd, Object... args)
			throws IOException {
		OneResponse res = super.call(cmd, args);
		if (res.isError())
			throw new IOException(res.getErrorMessage());

		return res;
	}

	public ImagePool imagePool(int filter) throws IOException {
		return (ImagePool) jaxbCall("imagepool.info", ImagePool.class, filter);
	}

	public HostPool hostPool() throws IOException {
		return (HostPool) jaxbCall("hostpool.info", HostPool.class);
	}

	public ClusterPool clusterPool() throws IOException {
		return (ClusterPool) jaxbCall("clusterpool.info", ClusterPool.class);
	}

	public UserPool userPool() throws IOException {
		return (UserPool) jaxbCall("userpool.info", UserPool.class);
	}

	public VirtualMachinePool vmPool(int filter) throws IOException {
		return (VirtualMachinePool) jaxbCall("vmpool.info",
				VirtualMachinePool.class, filter);
	}

	public VirtualNetworkPool vnPool(int filter) throws IOException {
		return (VirtualNetworkPool) jaxbCall("vnpool.info",
				VirtualNetworkPool.class, filter);
	}

	public int vmAllocate(String template) throws IOException {
		return jaxbCall("vm.allocate", template).getReturnCode();
	}

	public int clusterAllocate(String name) throws IOException {
		return jaxbCall("cluster.allocate", name).getReturnCode();
	}

	public int imageAllocate(String template) throws IOException {
		return jaxbCall("image.allocate", template).getReturnCode();
	}

	public void vmDeploy(int vmId, int hostId) throws IOException {
		jaxbCall("vm.deploy", vmId, hostId);
	}

	public void vmAction(VmAction action, long l) throws IOException {
		jaxbCall("vm.action", action.getCode(), (int) l);
	}

	public void vmMigrate(int vmId, int hostId, boolean isLiveMigrate)
			throws IOException {
		jaxbCall("vm.migrate", vmId, hostId, isLiveMigrate);
	}

	public VirtualMachine vmInfo(long vmId) throws IOException {
		return (VirtualMachine) jaxbCall("vm.info", VirtualMachine.class,
				(int) vmId);
	}

	public Cluster clusterInfo(int id) throws IOException {
		return (Cluster) jaxbCall("cluster.info", Cluster.class, id);
	}

	public int hostAllocate(String hostname, String imMadName,
			String vmmMadName, String transferManager, boolean isManaged)
			throws IOException {
		return jaxbCall("host.allocate", hostname, imMadName, vmmMadName,
				transferManager, isManaged).getReturnCode();
	}

	public Host hostInfo(int hostId) throws IOException {
		return (Host) jaxbCall("host.info", Host.class, hostId);
	}

	public ImageDetails imageInfo(int imageId) throws IOException {
		return (ImageDetails) jaxbCall("image.info", ImageDetails.class,
				imageId);
	}

	public void hostDelete(int hostId) throws IOException {
		jaxbCall("host.delete", hostId);
	}

	public void imageDelete(int imageId) throws IOException {
		jaxbCall("image.delete", imageId);
	}

	public void clusterDelete(int clusterId) throws IOException {
		jaxbCall("cluster.delete", clusterId);
	}

	public void clusterAdd(int hostId, int clusterId) throws IOException {
		jaxbCall("cluster.add", hostId, clusterId);
	}

	public void clusterRemove(int hostId) throws IOException {
		jaxbCall("cluster.remove", hostId);
	}

	public int imageUpdate(int imageId, String attributeName, String newValue)
			throws IOException {
		return jaxbCall("image.update", imageId, attributeName, newValue)
				.getReturnCode();
	}

	public int imageRemoveAttribute(int imageId, String attributeName)
			throws IOException {
		return jaxbCall("image.rmatt", imageId, attributeName).getReturnCode();
	}

	public void imageEnable(int imageId, boolean enable) throws IOException {
		jaxbCall("image.enable", imageId, enable);
	}

	public void imagePublish(int imageId, boolean publish) throws IOException {
		jaxbCall("image.publish", imageId, publish);
	}

	public void hostEnable(int hostId, boolean enable) throws IOException {
		jaxbCall("host.enable", hostId, enable);
	}

	public int vnAllocate(String template) throws IOException {
		return jaxbCall("vn.allocate", template).getReturnCode();
	}

	// VirtualNetworkDetails
	public VirtualNetworkDetails vnInfo(int vnId) throws IOException {
		return (VirtualNetworkDetails) jaxbCall("vn.info",
				VirtualNetworkDetails.class, vnId);
	}

	public void vnDelete(int vnId) throws IOException {
		jaxbCall("vn.delete", vnId);
	}

	public int userAllocate(String username, String password)
			throws IOException, NoSuchAlgorithmException {
		return jaxbCall("user.allocate", username, encodeSHA1(password))
				.getReturnCode();
	}

	public void userPasswd(int uid, String password) throws IOException,
			NoSuchAlgorithmException {
		jaxbCall("user.passwd", uid, encodeSHA1(password));
	}

	/**
	 * @param password
	 * @return
	 * @throws NoSuchAlgorithmException
	 */
	private String encodeSHA1(String password) throws NoSuchAlgorithmException {

		MessageDigest md = MessageDigest.getInstance("SHA-1");
		byte[] digest = md.digest(password.getBytes());

		String hash = "";

		for (byte aux : digest) {
			int b = aux & 0xff;

			if (Integer.toHexString(b).length() == 1) {
				hash += "0";
			}

			hash += Integer.toHexString(b);
		}
		return hash;
	}

	/**
	 * Returns the User object for a user id. 
	 * This method can only be called by an OpenNebula admin user.
	 * @param userId
	 * @return The User for user id userId
	 * @throws IOException if a communication problem with the server occured, or if called by a user who is not an OpenNebula admin.
	 */
	public User userInfo(int userId) throws IOException {
		for (User user : userPool().getUsers()) {
			if (user.getId() == userId)
				return user;
		}
		return null;
		// return (User) jaxbCall("user.info", User.class, userId);
	}

	public void userDelete(int userId) throws IOException {
		jaxbCall("user.delete", userId);
	}

	/**
	 * @param uid
	 * @return
	 * @throws IOException
	 */
	public String getUserName(int uid) throws IOException {
		return userInfo(uid).getName();
	}
}
